/**
 * \file circ_common.c
 * \brief these functions are called by defaut if there is no special implementation for a specific circuit
 * \version $Revision:$
 * \date Last change: $Date:$
 * \author Oliver Dawid
 * \author Last change by: $Author:$
 */
/*----------------------------------------------------------------------------
 *  circd server for Linux
 *
 *  Copyright (c) 2005-2016 - Oliver Dawid, fli4l-Team <team@fli4l.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *----------------------------------------------------------------------------
 */

#include "circd.h"

type_functable_t unknown_tab = {
	NULL, /* init */
	NULL, /* start */
	NULL, /* dial */
	NULL, /* hangup */
	NULL, /* stop */
	NULL, /* pause */
	NULL, /* cont */
	NULL, /* dialmode */
	NULL, /* hup_timeout */
	NULL, /* status */
	NULL, /* calc_rates */
	NULL, /* calc_iobytes */
	NULL, /* read_persistence */
	NULL, /* write_persistence */
};

static void cb_system_data(int fd, void *context);
static void async_system(char *cmd, request_t *rp);

char cmd_buf[1024];

/// initialize the circuit
void circ_init(circuit_t *circ)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl init %s", circ->type_str, circ->name);
	async_system(cmd_buf, NULL);
}

/// activate circuit for use
void circ_start(request_t *rp)
{
	/* script must return device Syntax: OK: DEVICE or FAIL*/
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl start %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// dial in
void circ_dial(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl dial %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// hang up
void circ_hangup(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl hangup %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// deactivate circuit
void circ_stop(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl stop %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// pause the circuit - no traffic is sent in the paused time but connection is kept
void circ_pause(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl pause %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// continue the circuit
void circ_cont(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl cont %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// read/ switch dialmode
void circ_dialmode(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl dialmode %s %s", rp->circ->type_str, dialmode_str[rp->data.mode], rp->circ->name);
	async_system(cmd_buf, rp);
}

/// read/ change hangup timeout
void circ_hup_timeout(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl hup_timeout %d %s", rp->circ->type_str, (int)rp->data.hup_timeout, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// return status of this circuit
void circ_status(request_t *rp)
{
	snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl status %s", rp->circ->type_str, rp->circ->name);
	async_system(cmd_buf, rp);
}

/// calculate throughput for last interval
void circ_calc_troughput(circuit_t *circ)
{
	
}

/// read statistics for this circuit from file
void circ_read_persistence(circuit_t *circ)
{
	/* read timetable charge */
}

/// write statistics to file
void circ_write_persistence(circuit_t *circ)
{
	/* write timetable charge */
}

/// systemcall handled asynchronously to avoid blocking circd
static void async_system(char *cmd, request_t *rp)
{
	int fds[2], ret;
	pid_t pid;
	
	if (rp) {
		pipe(fds);
	} else {
		fds[0] = open("/dev/null", O_RDONLY);
		fds[1] = open("/dev/null", O_WRONLY);
	}
	if ((pid = fork()) == 0) {
		circd_syslog(LOG_INFO, "Executing %s", cmd);
		close(1);
		close(2);
		close(fds[0]);
		dup2(fds[1], 1);
		dup2(fds[1], 2);
		close(fds[1]);
		ret = (int)((unsigned int) system(cmd) >> 8);
		exit(ret);
	} else if (pid > 0) {
		close(fds[1]);
		if (rp)
			register_event(fds[0], POLLIN, cb_system_data, rp);
		else
			close(fds[0]);
	} else {
		close(fds[0]);
		close(fds[1]);
	}
}

/// callback function called, when forked process from async_system returns any data
static void cb_system_data(int fd, void *context)
{
	struct _action_str_t *ap = action_str;
	char buf[1024], *p = buf;
	ssize_t len;
	request_t *rp = (request_t *)context;
	int success = 0;
	size_t count = 5;
	struct iovec v[6];
	
	unregister_event(fd, POLLIN);
	len = read(fd, &buf, sizeof(buf));
	if (len > 2 && !memcmp(buf, "OK", 2))
		success = 1;
	if (rp->out_fd > -1) {
		while (ap && ap->action != rp->cb)
			++ap;
		if (ap) {
			v[0].iov_base = ap->name;
			v[0].iov_len = strlen(ap->name);
		} else {
			v[0].iov_base = "Unknown command";
			v[0].iov_len = 15;
		}
		v[1].iov_base = " ";
		v[1].iov_len = 1;
		v[2].iov_base = rp->circ->name;
		v[2].iov_len = strlen(rp->circ->name);
		if (success) {
			v[3].iov_base = ": OK: ";
			v[3].iov_len = 6;
		} else {
			v[3].iov_base = ": FAIL: ";
			v[3].iov_len = 8;
		}
		if (len) {
			++count;
			v[4].iov_base = p;
			v[4].iov_len = len;
			write(rp->out_fd, p, len);
		}
		v[count - 1].iov_base = "\n";
		v[count - 1].iov_len = 1;
		writev(rp->out_fd, v, count);
	}
	if (rp->cb)
		rp->cb(success, buf, rp->context);
	remove_request(rp);
	free(rp);
}

