/**
 * \file circuit.c
 * \brief circuit control for non-lcr and lcr circuits
 * \version $Revision:$
 * \date Last change: $Date:$
 * \author Oliver Dawid
 * \author Last change by: $Author:$
 */

/*----------------------------------------------------------------------------
 *  circd server for Linux
 *
 *  Copyright (c) 2005-2016 - Oliver Dawid, fli4l-Team <team@fli4l.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *----------------------------------------------------------------------------
 */

#include "circd.h"

struct _tables_t {
	type_functable_t *tab;
};

struct _tables_t tables[] = {
	{ &pppoe_tab },
	{ &ippp_tab },
	{ &isdn_tab },
	{ &pptp_tab },
	{ &ppp_tab },
	{ &eth_tab },
	{ &vpn_tab },
	{ &unknown_tab },
	{ NULL }
};

static request_t *create_request(circuit_t *circ, circuit_func func, int out_fd, request_data_t *data, time_t request_timeout);
static request_t *find_request(circuit_t *circ, circuit_func func);
static void remove_all_requests(circuit_t *circ);
static void add_request(circuit_t *circ, void *action, time_t timeout);

static void cb_lcr_timer(void *context);
static void cb_nonlcr_timer(void *context);
static void cb_stop_circuit(int success, char *answer, void *context);
static void cb_start_circuit(int success, char *answer, void *context);
static void cb_start_lcr_circuit(int success, char *answer, void *context);

circuit_t *find_circuit(char *name)
{
	list_head *le;
	circuit_t *cp;
	
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		if (!strcmp(cp->name, name))
			return cp;
	}
	return NULL;
}

circuit_t *find_lcr_circuit(int day, int hour, int *tab_index)
{
	list_head *le;
	circuit_t *cp;
	int i, bit;
	time_table_t *tt;
	
	bit = day * HOURS_PER_DAY + hour;
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		if (cp->lcr) {
			tt = cp->time_tab;
			for (i = 0; i < cp->time_tab_count; ++i, ++tt) {
				if (isset(tt->tab, bit)) {
					*tab_index = i;
					return cp;
				}
			}
		}
	}
	return NULL;
}

void check_nonlcr_circuits(void)
{
	list_head *le;
	circuit_t *cp;
	int i;
	time_table_t *tt;
	int bit;
	request_t *n;
	struct tm *tm;
	time_t seconds;
	
	seconds = time((time_t *) 0);
	tm = localtime(&seconds);
	if (is_holiday(tm->tm_mday, tm->tm_mon + 1, tm->tm_year +1900))
		bit = SUNDAY * HOURS_PER_DAY + tm->tm_hour;
	else
		bit = tm->tm_wday * HOURS_PER_DAY + tm->tm_hour;
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		if (!cp->lcr) {
			tt = cp->time_tab;
			for (i = 0; i < cp->time_tab_count; ++i, ++tt) {
				if (isset(tt->tab, bit) && cp->current_tab == i) {
					/* only change something if timetable has changed */
					remove_all_requests(cp);
					if (tt->u.mode == cm_disable && cp->current_status != cs_disabled) {
						add_request(cp, circuit_stop, 1);
					} else if (tt->u.mode == cm_start && cp->current_status == cs_disabled) {
						add_request(cp, circuit_start, cp->start_timeout);
					} else if (tt->u.mode == cm_dial) {
						switch (cp->current_status) {
							case cs_disabled:
								add_request(cp, circuit_start, cp->start_timeout);
							case cs_starting:
							case cs_offline:
							case cs_hangingup:
								add_request(cp, circuit_dial, cp->dial_timeout);
								break;
							case cs_pausing:
								add_request(cp, circuit_cont, 1);
								break;
							default:
								break;
						}
					} else if (tt->u.mode == cm_pause) {
						switch (cp->current_status) {
							case cs_disabled:
								add_request(cp, circuit_start, cp->start_timeout);
							case cs_starting:
							case cs_offline:
							case cs_hangingup:
								add_request(cp, circuit_dial, cp->dial_timeout);
							case cs_online:
							case cs_dialing:
								add_request(cp, circuit_pause, cp->dial_timeout);
								break;
							default:
								break;
						}
					}
					if (tt->u.mode == cm_disable) {
						n = create_request(cp, circuit_stop, -1, NULL, 1);
						break;
					}
				}
				break;
			}
		}
	}
	set_timer(TIMEOUT_CHECK_LCR, cb_nonlcr_timer, NULL);
}

void check_lcr_circuits(void)
{
	circuit_t *new_circuit;
	int index;
	struct tm *tm;
	time_t seconds;
	int day;
	request_t *n1, *n2;
	request_data_t rd;
	static int print_lcr = 1;
	
	if (current_lcr_mode != cm_automatic && current_lcr_mode != cm_start)
		return;
	
	seconds = time((time_t *) 0);
	tm = localtime(&seconds);
	if (is_holiday(tm->tm_mday, tm->tm_mon + 1, tm->tm_year +1900))
		day = SUNDAY;
	else
		day = tm->tm_wday;
	new_circuit = find_lcr_circuit(day, tm->tm_hour, &index);
	if (print_lcr) {
		circd_syslog(LOG_INFO, "LCR Circuit %s", new_circuit? new_circuit->name: "-");
		print_lcr = 0;
	}
	if (new_circuit != current_circuit) {
		print_lcr = 1;
		circd_syslog(LOG_INFO, "Switch from circuit %s to %s", current_circuit? current_circuit->name: "-", new_circuit? new_circuit->name: "-");
		rd.mode = current_lcr_mode;
		if ((n1 = create_request(current_circuit, circuit_stop, -1, NULL, 1)) != NULL && 
			(n2 = create_request(new_circuit, circuit_start, -1, &rd, new_circuit->start_timeout)) != NULL) {
			if (current_circuit) {
				if (!find_request(current_circuit, circuit_stop)) {
					n1->depend = n2; /* if this action fails, dependend request is deleted */
					n1->cb = cb_stop_circuit;
					list_add_tail(&n1->list, &current_circuit->pending_requests);
					circd_syslog(LOG_INFO, "Queue stop event for circuit %s", current_circuit->name);
				} else {
					circd_syslog(LOG_INFO, "Already shutting down %s", current_circuit->name);
					free(n1);
				}
			} else {
				circd_syslog(LOG_INFO, "No current circuit to shut down");
				free(n1);
			}
			if (new_circuit) {
				if (!find_request(new_circuit, circuit_start)) {
					n2->cb = cb_start_lcr_circuit;
					n2->context = new_circuit;
					list_add_tail(&n2->list, &new_circuit->pending_requests);
					circd_syslog(LOG_INFO, "Queue start event for circuit %s", new_circuit->name);
				} else {
					circd_syslog(LOG_INFO, "Already starting %s", new_circuit->name);
					free(n2);
				}
			} else {
				circd_syslog(LOG_INFO, "No circuit to start");
				free(n2);
			}
		} else {
			circd_syslog(LOG_WARNING, "Unable to create requests: %s", strerror(errno));
		}
	}
	set_timer(TIMEOUT_CHECK_LCR, cb_lcr_timer, NULL);
}

static void cb_lcr_timer(void *context)
{
	check_lcr_circuits();
}

static void cb_nonlcr_timer(void *context)
{
	check_nonlcr_circuits();
}

static void cb_stop_circuit(int success, char *answer, void *context)
{
	circuit_t *circ = (circuit_t *)context;
	device_t **d1 = &circ->dev, *d2;

	if (success) {
		circ->current_status = cs_disabled;
		/* clear dependencies */
		while (*d1) {
			d2 = (*d1)->next;
			*d1 = NULL;
			d1 = &d2->next;
		}
	} else {
	}
}

static void cb_start_circuit(int success, char *answer, void *context)
{
	circuit_t *circ = (circuit_t *)context;
	
	if (success) {
		circ->current_status = cs_offline;
		/* TODO: we should check the answer to be some kind of device here or in create_device */
		if ((circ->dev = find_device(answer)) == NULL) {
			circ->dev = create_device(answer);
		} else {
			reset_device_accounting(circ->dev);
		}
	}
}

static void cb_start_lcr_circuit(int success, char *answer, void *context)
{
	circuit_t *circ = (circuit_t *)context;
	
	if (success) {
		current_circuit = circ;
		circ->current_status = cs_offline;
		circ->dev = find_device(answer);
		/* TODO: we should check the answer to be some kind of device here or in create_device */
		if ((circ->dev = find_device(answer)) == NULL) {
			circ->dev = create_device(answer);
		} else {
			reset_device_accounting(circ->dev);
		}
	} else {
		current_circuit = NULL;
	}
}

void circuits_calc_rates(void)
{
	list_head *le;
	circuit_t *cp;
	
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		if (tables[cp->type].tab->calc_rates)
			tables[cp->type].tab->calc_rates(cp);
		else
			circ_calc_troughput(cp);
	}
}

void circuit_init(circuit_t *circ)
{
	if (tables[circ->type].tab->init)
		tables[circ->type].tab->init(circ);
	else
		circ_init(circ);
}

void circuit_start(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->start)
		tables[circ->type].tab->start(rp);
	else
		circ_start(rp);
}

void circuit_dial(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->dial)
		tables[circ->type].tab->dial(rp);
	else
		circ_dial(rp);
}

void circuit_hangup(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->hangup)
		tables[circ->type].tab->hangup(rp);
	else
		circ_hangup(rp);
}

void circuit_stop(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->stop)
		tables[circ->type].tab->stop(rp);
	else
		circ_stop(rp);
}

void circuit_pause(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->pause)
		tables[circ->type].tab->pause(rp);
	else
		circ_pause(rp);
}

void circuit_cont(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->cont)
		tables[circ->type].tab->cont(rp);
	else
		circ_cont(rp);
}

void circuit_dialmode(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->dialmode)
		return tables[circ->type].tab->dialmode(rp, rp->data.mode);
	else
		return circ_dialmode(rp);
}

void circuit_hup_timeout(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->hup_timeout)
		return tables[circ->type].tab->hup_timeout(rp, rp->data.hup_timeout);
	else
		return circ_hup_timeout(rp);
}

void circuit_status(request_t *rp)
{
	circuit_t *circ = rp->circ;

	if (tables[circ->type].tab->status)
		tables[circ->type].tab->status(rp);
	else
		circ_status(rp);
}

void circuit_read_persistence(circuit_t *circ)
{
	if (tables[circ->type].tab->read_persistence)
		tables[circ->type].tab->read_persistence(circ);
	else
		circ_read_persistence(circ);
}

void circuit_write_persistence(circuit_t *circ)
{
	if (tables[circ->type].tab->write_persistence)
		tables[circ->type].tab->write_persistence(circ);
	else
		circ_write_persistence(circ);
}

static request_t *create_request(circuit_t *circ, circuit_func func, int out_fd, request_data_t *rd, time_t request_timeout)
{
	request_t *new_request;
	
	if ((new_request = (request_t *)calloc(1, sizeof(*new_request))) == NULL) {
		return NULL;
	}
	new_request->func = func;
	new_request->out_fd = out_fd;
	new_request->circ = circ;
	new_request->timeout = time(NULL) + request_timeout;
	if (rd)
		memcpy(&new_request->data, rd, sizeof(*rd));
	new_request->state = rs_queued;
	
	return new_request;
}

static request_t *find_request(circuit_t *circ, circuit_func func)
{
	list_head *le;
	request_t *rp;
	
	list_for_each(le, &circ->pending_requests) {
		rp = list_entry(le, request_t, list);
		if (rp->func == func && rp->circ == circ)
			return rp;
	}
	return NULL;	
}

static void remove_all_requests(circuit_t *circ)
{
	request_t *n;
	circuit_t *cp;
	
	if ((n = find_request(cp, circuit_start)) != NULL) {
		/* TODO: remove timer entry for this request */
		list_del(&n->list);
		free(n);
	}
	if ((n = find_request(cp, circuit_dial)) != NULL) {
		/* TODO: remove timer entry for this request */
		list_del(&n->list);
		free(n);
	}
	if ((n = find_request(cp, circuit_pause)) != NULL) {
		/* TODO: remove timer entry for this request */
		list_del(&n->list);
		free(n);
	}
	if ((n = find_request(cp, circuit_cont)) != NULL) {
		/* TODO: remove timer entry for this request */
		list_del(&n->list);
		free(n);
	}
	if ((n = find_request(cp, circuit_stop)) != NULL) {
		/* TODO: remove timer entry for this request */
		list_del(&n->list);
		free(n);
	}
}

void check_for_requests(void)
{
	request_t *rp;
	circuit_t *cp;
	list_head *le;
	
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		if (list_empty(&cp->pending_requests))
			continue;
	
		circd_syslog(LOG_INFO, "Found request for %s: ", cp->name);
		rp = list_entry(cp->pending_requests.next, request_t, list);
		if (rp->state == rs_pending) {
			circd_syslog(LOG_INFO, " State: pending");
			continue;
		}
		
		circd_syslog(LOG_INFO, " Set state to pending and call function");
		rp->state = rs_pending;
		rp->func(rp);
	}
}

void remove_request(request_t *rp)
{
	list_del(&rp->list);
}

static void add_request(circuit_t *circ, void *action, time_t timeout)
{
	request_t *n;
	
	if (!find_request(circ, action)) {
		n = create_request(circ, action, -1, NULL, timeout);
		list_add_tail(&n->list, &circ->pending_requests);
	}
}
