/**
 * \file config.c
 * \brief functions to read configuration files
 * \version $Revision:$
 * \date Last change: $Date:$
 * \author Oliver Dawid
 * \author Last change by: $Author:$
 */

/*----------------------------------------------------------------------------
 *  circd server for Linux
 *
 *  Copyright (c) 2005-2016 - Oliver Dawid, fli4l-Team <team@fli4l.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *----------------------------------------------------------------------------
 */

#include "circd.h"

typedef struct {
	int start_day;   // entry start day
	int end_day;     // entry end day
	int start_hour;  // entry start time
	int end_hour;    // entry end time
	int hup_timeout; // timeout of entry
	int lcr;         // 0 if non-lcr, otherwise 1
	union {
		circ_mode_t mode;    // mode of circuit during activated time
		circ_lcrflag_t flag; // flags for lcr-circuit during activated time
	} u;
	double charge_p_int; // charge per interval
	int charge_int;      // charge interval (in seconds)
	circ_type_t type;    // circuit type - see also type_str
	char device[16];     // device used by this circuit - maybe empty
	char bw_buf[32];     // bandwidth threshold and time, where slave channel should be added
	char name_buf[16];   // name of circuit
	char phone_buf[32];  // phone number(s) - may be empty
	char type_str[16];   // circuit type string to make it easy to call a script for this type
} circuit_params_t;

static void read_conf_file(const char *path);
static int insert_circuit(circuit_params_t *circ_params);
static void fill_time_table(int start_day, int end_day, int start_hour, int end_hour, time_table_t *tt);
static int scan_line(char *line, circuit_params_t *circ_params);
static int scan_day(char **pp, int *res);
static int scan_hour(char **pp, int *res);
static int scan_string(char **pp, char *dest, int len);
static int scan_number(char **pp, int *res);
static int scan_double(char **pp, double *res);
static int scan_yesno(char **pp, int *res);
static int scan_mode(char **pp, circ_mode_t *res);
static int scan_lcr_flag(char **pp, circ_lcrflag_t *res);
static int scan_type(char **pp, circ_type_t *res, char *type_str, int len);
static void *my_realloc(void *p, size_t size, size_t new_size);

int
read_config(void)
{
	glob_t g;
	int i;
	struct stat st;
#ifdef DEBUG
	list_head *le;
	circuit_t *cp;
#endif
	memset(&g, 0, sizeof(g));
	if (glob("./etc/circd.d/*", GLOB_NOSORT, NULL, &g) == 0) {
		for (i = 0; i < g.gl_pathc; ++i) {
			if (stat(g.gl_pathv[i], &st) == 0 && S_ISREG(st.st_mode))
				read_conf_file(g.gl_pathv[i]);
		}
	}
#ifdef DEBUG
	list_for_each(le, &circuits) {
		cp = list_entry(le, circuit_t, list);
		circd_syslog(LOG_INFO, "%s\n", cp->name);
	}
#endif
	globfree(&g);
	
	return 0;
}

static void
read_conf_file(const char *path)
{
	FILE *fp;
	char buf[512];
	int i = 0;
	circuit_params_t circ_params;
	char* p;
	
	if ((fp = fopen(path, "r")) != NULL) {
		while (fgets(buf, sizeof(buf), fp)) {
			++i;
			if (*buf == '#')
				continue;
			if (scan_line(buf, &circ_params) == -1) {
				circd_syslog(LOG_WARNING, "Line %d in config file %s is invalid", i, path);
				continue;
			}
#ifdef DEBUG
			sprintf(buf, "s_d: %d e_d: %d s_h: %2d e_h: %2d hup_t: %4d lcr: %d ", circ_params.start_day, 
					circ_params.end_day, circ_params.start_hour, circ_params.end_hour,
				       	circ_params.hup_timeout, circ_params.lcr);
			if (circ_params.lcr) {
				strcat(buf, "flag: ");
				strcat(buf, lcrflag_str[circ_params.u.flag]);
			} else {
				strcat(buf, "mode: ");
				strcat(buf, circmode_str[circ_params.u.mode]);
			}
			p = buf + strlen(buf);
			sprintf(p, " ch_pt: %1.3f ch_t: %d tp: %d dev: %s bw: %s nm: %s ph: %s", circ_params.charge_p_int,
				       circ_params.charge_int, circ_params.type, circ_params.device, circ_params.bw_buf,
				       circ_params.name_buf, circ_params.phone_buf);
			circd_syslog(LOG_INFO, buf);
#endif
			if (insert_circuit(&circ_params) == -1) {
				fclose(fp);
				return;
			}
		}
		fclose(fp);
	} else
		circd_syslog(LOG_ERR, "Could not open config file '%s': %s", path, strerror(errno));
}

static int 
insert_circuit(circuit_params_t *circ_params)
{
	circuit_t *new_circuit, *nc;
	int size, n, delta;
	char *p, *p2;
	time_table_t *tt;
	char cmd_buf[256];
	struct stat st;

	if (circ_params->device[0] == '-')
		circ_params->device[0] = '\0';
	if (circ_params->phone_buf[0] == '-')
		circ_params->phone_buf[0] = '\0';
	if (circ_params->type == ct_unknown) {
		snprintf(cmd_buf, sizeof(cmd_buf), EXEC_PREFIX "%s-circuit-ctrl", circ_params->type_str);
		if (stat(cmd_buf, &st) || !(st.st_mode & S_IFREG) || !(st.st_mode & S_IXUSR)) {
			circd_syslog(LOG_ERR, "Cannot find executable control file: %s", cmd_buf);
			return -1;
		}
	}
	
	if ((new_circuit = find_circuit(circ_params->name_buf)) == NULL) {
		size = sizeof(*new_circuit) + strlen(circ_params->name_buf) + strlen(circ_params->device) + strlen(circ_params->phone_buf) + 3/* null termination for strings */;
		size = (size + 7) & ~7; /* align to 8 byte */
		size += sizeof(time_table_t);
		if ((new_circuit = (circuit_t *)calloc(1, size)) == NULL) {
			circd_syslog(LOG_ERR, "calloc: Could not alloc memory: %s", strerror(errno));
			return -1;
		}
		new_circuit->size = size;
		p = (char *)new_circuit + sizeof(*new_circuit);
		/* no need to use strlcpy because buffers are all terminated and memory is big enough */
		strcpy(p, circ_params->name_buf);
		new_circuit->name = p;
		p += strlen(circ_params->name_buf) + 1;
		if (circ_params->device[0]) {
			strcpy(p, circ_params->device);
			new_circuit->device = p;
			p += strlen(circ_params->device) + 1;
		}
		if (circ_params->phone_buf[0]) {
			strcpy(p, circ_params->phone_buf);
			if ((p2 = strchr(p, ':')) != NULL) {
				*p2++ = '\0';
				if (*p2)
					new_circuit->phone_in = p2;
			}
			if (*p)
				new_circuit->phone_out = p;
			p += strlen(circ_params->phone_buf) + 1;
		}
		if ((n = atoi(circ_params->bw_buf)) != 0) {
			new_circuit->bandwidth_rate = n;
			if ((p2 = strchr(circ_params->bw_buf, ':')) != NULL)
				n = atoi(++p2);
			else
				n = -1;
			new_circuit->bandwidth_time = n;
		} else
			new_circuit->bandwidth_rate = -1;
		p = (char *)(((u_int32_t)p + 7) & ~7); /* align timetable to 8 byte */
		new_circuit->time_tab = (time_table_t *)p;
		new_circuit->time_tab_count = 1;
		fill_time_table(circ_params->start_day, circ_params->end_day, circ_params->start_hour, circ_params->end_hour, new_circuit->time_tab);
		new_circuit->time_tab->charge_int = circ_params->charge_int;
		new_circuit->time_tab->charge_p_int = circ_params->charge_p_int;
		new_circuit->lcr = circ_params->lcr;
		if (new_circuit->lcr)
			new_circuit->time_tab->u.flag = circ_params->u.flag;
		else
			new_circuit->time_tab->u.mode = circ_params->u.mode;
		new_circuit->type = circ_params->type;
		new_circuit->hup_timeout = circ_params->hup_timeout;
		new_circuit->current_tab = -1;
		memcpy(new_circuit->type_str, circ_params->type_str, MIN(sizeof(new_circuit->type_str), sizeof(circ_params->type_str)));
		new_circuit->type_str[sizeof(new_circuit->type_str) - 1] = '\0';
		list_init(&new_circuit->pending_requests);
	} else {
		size = new_circuit->size + sizeof(time_table_t);
		list_del(&new_circuit->list);
		if ((nc = (circuit_t *)my_realloc(new_circuit, new_circuit->size, size)) == NULL) {
			circd_syslog(LOG_ERR, "calloc: Could not alloc memory: %s", strerror(errno));
			list_add_tail(&new_circuit->list, &circuits);
			return -1;
		}
		/* correct references */
		delta = (char *)nc - (char *)new_circuit;
		nc->name += delta;
		if (nc->device)
			nc->device += delta;
		if (nc->phone_out)
			nc->phone_out += delta;
		if (nc->phone_in)
			nc->phone_in += delta;
		nc->time_tab = (time_table_t *)((char *)nc->time_tab + delta);
		list_init(&nc->pending_requests);
		new_circuit = nc;
		tt = new_circuit->time_tab + new_circuit->time_tab_count;
		memset(tt, 0, sizeof(*tt));
		fill_time_table(circ_params->start_day, circ_params->end_day, circ_params->start_hour, circ_params->end_hour, tt);
		tt->charge_int = circ_params->charge_int;
		tt->charge_p_int = circ_params->charge_p_int;
		if (new_circuit->lcr)
			tt->u.flag = circ_params->u.flag;
		else
			tt->u.mode = circ_params->u.mode;
		++new_circuit->time_tab_count;
	}
	
	list_add_tail(&new_circuit->list, &circuits);
	return 0;
}

/* fill the time table: 
* start_day may be == end_day but start_hour must be != end_hour
* using modulo operations to get rid of distinction if start_hour > end_hour etc.
* so just add DAYS_PER_WEEK to end_day and HOURS_PER_DAY to end_hour if necessary 
*/
static void fill_time_table(int start_day, int end_day, int start_hour, int end_hour, time_table_t *tt)
{
	int i, j;
	char buf[64];
	char* p;
	if (end_day < start_day)
		end_day += DAYS_PER_WEEK;
	if (end_hour < start_hour)
		end_hour += HOURS_PER_DAY;
	for (i = start_day; i <= end_day; ++i) {
		for (j = start_hour; j < end_hour; ++j) {
			setbit(tt->tab, (i % DAYS_PER_WEEK)* HOURS_PER_DAY + (j % HOURS_PER_DAY));
			//circd_syslog(LOG_INFO, "d %d h %d (%d)\n", i, j, (i % DAYS_PER_WEEK)* HOURS_PER_DAY + (j % HOURS_PER_DAY));
		}
	}

	for (i = 0; i <= 6; ++i) {
		sprintf(buf, "Day %d: ", i);
		p = buf + strlen(buf);
		for (j = 0; j < 24; ++j) {
			if (isset(tt->tab, (i * 24) + j))
				sprintf(p, "1 ");
			else
				sprintf(p, "0 ");
			p += 2;
		}
		circd_syslog(LOG_INFO, buf);
	}
}

static int scan_line(char *line, circuit_params_t *circ_params)
{
	char *p = line;
	memset(circ_params, 0, sizeof(*circ_params));
	
	if (scan_day(&p, &circ_params->start_day) == -1)
		return -1;
	if (*p != '-')
		return -1;
	++p;
	if (scan_day(&p, &circ_params->end_day) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_hour(&p, &circ_params->start_hour) == -1)
		return -1;
	if (*p != '-')
		return -1;
	++p;
	if (scan_hour(&p, &circ_params->end_hour) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_string(&p, circ_params->device, sizeof(circ_params->device)) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_string(&p, circ_params->bw_buf, sizeof(circ_params->bw_buf)) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_number(&p, &circ_params->hup_timeout) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_yesno(&p, &circ_params->lcr) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (circ_params->lcr) {
		if (scan_lcr_flag(&p, &circ_params->u.flag) == -1)
			return -1;
	} else {
		if (scan_mode(&p, &circ_params->u.mode) == -1)
			return -1;
	}
	if (*p != ' ')
		return -1;
	++p;
	if (scan_string(&p, circ_params->phone_buf, sizeof(circ_params->phone_buf)) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_string(&p, circ_params->name_buf, sizeof(circ_params->name_buf)) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_double(&p, &circ_params->charge_p_int) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_number(&p, &circ_params->charge_int) == -1)
		return -1;
	if (*p != ' ')
		return -1;
	++p;
	if (scan_type(&p, &circ_params->type, circ_params->type_str, sizeof(circ_params->type_str)) == -1)
		return -1;
	return 0;
}

static int scan_day(char **pp, int *res)
{
	if (!memcmp(*pp, "Mo", 2))
		*res = 1;
	else if (!memcmp(*pp, "Tu", 2))
		*res = 2;
	else if (!memcmp(*pp, "We", 2))
		*res = 3;
	else if (!memcmp(*pp, "Th", 2))
		*res = 4;
	else if (!memcmp(*pp, "Fr", 2))
		*res = 5;
	else if (!memcmp(*pp, "Sa", 2))
		*res = 6;
	else if (!memcmp(*pp, "Su", 2))
		*res = 0;
	else 
		return -1;
	*pp += 2;
	return 0;
}

static int scan_hour(char **pp, int *res)
{
	*res = atoi(*pp);
	
	if (*res >= 0 && *res <= 24) {
		*pp += 2;
		return 0;
	} else
		return -1;
}

static int scan_string(char **pp, char *dest, int len)
{
	char *p = *pp;
	while(*p != ' ' && *p != '\0' && len > 1) {
		*dest++ = *p++;
		--len;
	}
	*pp = p;
	*dest = '\0';

	return 0;
}

static int scan_number(char **pp, int *res)
{
	char *p;
	*res = strtol(*pp, &p, 0);
	*pp = p;

	return 0;
}

static int scan_double(char **pp, double *res)
{
	char *p;
	*res = strtod(*pp, &p);
	*pp = p;
	
	return 0;
}

static int scan_yesno(char **pp, int *res)
{
	if (!memcmp(*pp, "yes", 3)) {
		*res = 1;
		*pp += 3;
	} else if (!memcmp(*pp, "no", 2)) {
		*res = 0;
		*pp += 2;
	} else
		return -1;
	return 0;
}

static int scan_mode(char **pp, circ_mode_t *res)
{
	if (!memcmp(*pp, "disable", 7)) {
		*res = cm_disable;
		*pp += 7;
	} else if (!memcmp(*pp, "start", 5)) {
		*res = cm_start;
		*pp += 5;
	} else if (!memcmp(*pp, "dial", 4)) {
		*res = cm_dial;
		*pp += 4;
	} else if (!memcmp(*pp, "pause", 5)) {
		*res = cm_pause;
		*pp += 5;
	} else
		return -1;
	return 0;
}

static int scan_lcr_flag(char **pp, circ_lcrflag_t *res)
{
	if (!memcmp(*pp, "yes", 3)) {
		*res = cf_active;
		*pp += 3;
	} else if (!memcmp(*pp, "no", 2)) {
		*res = cf_enable;
		*pp += 2;
	} else if (!memcmp(*pp, "disable", 7)) {
		*res = cf_disable;
		*pp += 7;
	} else
		return -1;
	return 0;
}

static int scan_type(char **pp, circ_type_t *res, char *type_str, int len)
{
	char *p;
	
	if (!memcmp(*pp, "pppoe", 5))
		*res = ct_pppoe;
	else if (!memcmp(*pp, "ippp", 4))
		*res = ct_ippp;
	else if (!memcmp(*pp, "isdn", 4))
		*res = ct_isdn;
	else if (!memcmp(*pp, "pptp", 4))
		*res = ct_pptp;
	else if (!memcmp(*pp, "ppp", 3))
		*res = ct_ppp;
	else if (!memcmp(*pp, "eth", 3))
		*res = ct_eth;
	else if (!memcmp(*pp, "vpn", 3))
		*res = ct_vpn;
	else {
		*res = ct_unknown;
	}
	p = *pp;
	while(**pp >= 'a' && **pp <= 'z')
		++ *pp;
	if (*pp - p < len) {
		memcpy(type_str, p, *pp - p);
		type_str[*pp - p] = '\0';
	}

	return 0;
}

static void *my_realloc(void *p, size_t size, size_t new_size)
{
	char *newp;
	
	if ((newp = calloc(1, new_size)) != NULL) {
		memcpy(newp, p, size);
		memset(p, 0, size);
		free(p);
		return newp;
	} else {
		return NULL;
	}
}
