/**
 * \file events.c
 * \brief the main scheduler for timer and file descriptors
 * \version $Revision:$
 * \date Last change: $Date:$
 * \author Oliver Dawid
 * \author Last change by: $Author:$
 */

/*----------------------------------------------------------------------------
 *  circd server for Linux
 *
 *  Copyright (c) 2005-2016 - Oliver Dawid, fli4l-Team <team@fli4l.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *----------------------------------------------------------------------------
 */

#include "circd.h"

typedef struct {
	list_head list;
	u_int64_t time;
	timer_callback cb;
	void *context;
} timer_event_t;

static int rebuild_fdtab(void);
static void insert_timer(timer_event_t *new_tp);
void check_timer(void);
static u_int64_t get_usec_time(void);
static void dump_event_list(char *s);

static struct pollfd *fdtab = NULL;
static LIST_HEAD(unused_handler);
static LIST_HEAD(event_handler);
static LIST_HEAD(new_handler);
static LIST_HEAD(timer_events);
static int event_count = 0;
static int event_count_delta = 0;
static int rebuild = FALSE;

int register_event(int fd, int event, action_func af, void *context)
{
	list_head *le;
	event_handler_t *ehp;

	list_for_each(le, &event_handler) {
		/* try to find an existing active handler */
		ehp = list_entry(le, event_handler_t, list);
		if (ehp->fd == fd) {
			fdtab[ehp->tab_index].events |= event;
			break;
		}
	}
	if (le == &event_handler) {
		if (!list_empty(&unused_handler)) {
			/* reuse an unused handler */
			ehp = list_entry(unused_handler.next, event_handler_t, list);
			list_del(&ehp->list);
			memset(ehp, 0, sizeof(*ehp));
		} else {
			/* create a new handler */
			if ((ehp = (event_handler_t *)calloc(sizeof(*ehp), 1)) == NULL) {
				circd_syslog(LOG_ERR, "calloc failed: %s", strerror(errno));
				return -1;
			}
		}
		list_add_tail(&ehp->list, &new_handler);
		++event_count_delta;
		ehp->fd = fd;
		ehp->tab_index = -1;
		rebuild = TRUE;
		dump_event_list("register_event");
	}
	ehp->events |= event;
	if (event & POLLIN) {
		ehp->read_func = af;
		ehp->read_context = context;
	}
	if (event & POLLOUT) {
		ehp->write_func = af;
		ehp->write_context = context;
	}
	return 0;
}

int unregister_event(int fd, int event)
{
	list_head *le;
	event_handler_t *ehp;
	
	list_for_each(le, &event_handler) {
		ehp = list_entry(le, event_handler_t, list);
		if (ehp->fd == fd) {
			fdtab[ehp->tab_index].events &= ~event;
			ehp->events &= ~event;
			if (ehp->events == 0) {
				--event_count_delta;
				rebuild = TRUE;
			}
			dump_event_list("unregister_event");
			return 0;
		}
	} 
	return -1;
}

int poll_events(void)
{
	list_head *le;
	event_handler_t *ehp;
	int res, timeout;
	u_int64_t t;
	timer_event_t *tp;
	sigset_t sig_set;
	
	if (rebuild)
		rebuild_fdtab();
again:
	if (!list_empty(&timer_events)) {
		t = get_usec_time();
		tp = list_entry(timer_events.next, timer_event_t, list);
		timeout = tp->time - t;
		if (timeout < 1)
			timeout = 1;
	} else
		timeout = IMOND_TIMEOUT;
	/* allow all signals */
	sigemptyset(&sig_set);
	sigprocmask(SIG_SETMASK, &sig_set, NULL);
	if ((res = poll(fdtab, event_count, timeout)) == -1) {
		if (errno == EINTR) {
			/* some signal arrived and has been handled by signal handler */
			goto again;
		}
		circd_syslog(LOG_ERR, "poll: %m");
	}
	/* block all signals */
	sigfillset(&sig_set);
	sigprocmask(SIG_SETMASK, &sig_set, NULL);
	/* 1. check timer: lcr and non-lcr control is handled via timer */
	check_timer();
	/* 2. handle connections */
	if (res > 0) {
		list_for_each(le, &event_handler) {
			ehp = list_entry(le, event_handler_t, list);
			if (fdtab[ehp->tab_index].revents & (POLLIN | POLLERR | POLLHUP | POLLNVAL) && ehp->read_func) {
				ehp->read_func(ehp->fd, ehp->read_context);
			}
			if (fdtab[ehp->tab_index].revents & POLLOUT && ehp->write_func) {
				ehp->write_func(ehp->fd, ehp->write_context);
			}
		}
		if (rebuild)
			rebuild_fdtab();
	}
	/* 3. handle requests sequentially for each circuit */
	check_for_requests();

	return res;
}

int set_timer(int usecs, timer_callback cb, void *context)
{
	timer_event_t *tp;
	
	if ((tp = (timer_event_t *)calloc(1, sizeof(*tp))) == NULL) {
		return -1;
	}
	tp->time = get_usec_time() + (u_int64_t)usecs;
	tp->cb = cb;
	tp->context = context;
	insert_timer(tp);
	return 0;
}

void check_timer(void)
{
	timer_event_t *tp;
	u_int64_t t;
	
	t = get_usec_time();
	while(!list_empty(&timer_events)) {
		tp = list_entry(timer_events.next, timer_event_t, list);
		if (tp->time <= t) {
			list_del(&tp->list);
			tp->cb(tp->context);
			free(tp);
		} else {
			break;
		}
	}
}

void clear_timer(timer_callback cb, void *context)
{
	timer_event_t *tp;
	LIST_HEAD(tmp_list);
	
	while(!list_empty(&timer_events)) {
		tp = list_entry(&timer_events.next, timer_event_t, list);
		list_del(&tp->list);
		if (tp->cb == cb && tp->context == context) {
			free(tp);
		} else {
			list_add_tail(&tp->list, &tmp_list);
		}
	}
	if (!list_empty(&tmp_list)) {
		list_splice(&tmp_list, &timer_events);
	}
}

static void insert_timer(timer_event_t *new_tp)
{
	list_head *le;
	timer_event_t *tp = NULL;
	
	list_for_each(le, &timer_events) {
		tp = list_entry(le, timer_event_t, list);
		if (tp->time > new_tp->time)
			break;
	}
	if (tp)
		list_add(&new_tp->list, &tp->list);
	else
		list_add(&new_tp->list, &timer_events);
}
	
static int rebuild_fdtab(void)
{
	list_head *le;
	event_handler_t *ehp;
	int i = 0;
	struct pollfd *fds;
	
	rebuild = FALSE;

	list_splice(&new_handler, &event_handler);
	dump_event_list("rebuild_fdtab(1)");
	i = event_count + event_count_delta;
again:
	list_for_each(le, &event_handler) {
		ehp = list_entry(le, event_handler_t, list);
		if (!ehp->events) {
			list_del(&ehp->list);
			list_add_tail(&ehp->list, &unused_handler);
			goto again;
		}
	}
	if (i != event_count) {
		if (i > 0) {
			if ((fds = (struct pollfd *)realloc(fdtab, sizeof(*fds) * i)) == NULL) {
				circd_syslog(LOG_ERR, "realloc failed: %s", strerror(errno));
				/* we leave the delta as it is and hope for better times
				 * we use the table as it is so event_count remains the same
				 */
			} else {
				fdtab = fds;
				event_count_delta = 0;
				event_count = i;
			}
		} else {
			free(fdtab);
			fdtab = NULL;
		}
	}
	i = 0;
	list_for_each(le, &event_handler) {
		if (i == event_count) {
			/* can only happen, if realloc failed. so we handle up to event_count 
			 * fd's and keep the at the end of the event list
			 * TODO: maybe we should shut down those connections we cannot handle
			 * via poll()
			 */
			break;
		}
		ehp = list_entry(le, event_handler_t, list);
		fdtab[i].fd = ehp->fd;
		fdtab[i].events = ehp->events;
		ehp->tab_index = i;
		++i;
	}
	dump_event_list("rebuild_fdtab(2)");
	return 0;
}

static u_int64_t get_usec_time(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return (u_int64_t)tv.tv_sec * 1000 + (u_int64_t)tv.tv_usec / 1000;
}

/// function to debug fd rebuild process
static void dump_event_list(char *s)
{
	list_head *le;
	event_handler_t *ehp;
	int i = 0;
	circd_syslog(LOG_DEBUG, "%s:\nevent_count=%d event_count_delta=%d\n", s, event_count, event_count_delta);
	list_for_each(le, &event_handler) {
		ehp = list_entry(le, event_handler_t, list);
		circd_syslog(LOG_DEBUG, "%d: fd=%d tab_index=%d readfunc=0x%x writefunc=0x%x\n", i++, ehp->fd, ehp->tab_index, (unsigned int)ehp->read_func, (unsigned int)ehp->write_func);
	}
}
